#include <iomanip>
#include <string>
#include <algorithm>

#include <GL/glew.h>
#include <GL/freeglut.h>
#include <GL/gl.h>

#include <cuda_runtime.h>
#include <cuda_gl_interop.h>
#include <cuda_profiler_api.h>

#include <EGrabber.h>
#include "egrabber-cuda.h"

using namespace Euresys;

#define NB_BUFFERS 3

Options options;
Euresys::Sample::ConcurrencyLock lock;
Buffer *currentBuffer = NULL;
MyGrabber *grabber = NULL;

MyGrabber::MyGrabber(GenTL &gentl) : EGrabberCallbackMultiThread(gentl) {};

void MyGrabber::onNewBufferEvent(const NewBufferData &data) {
    Euresys::Sample::AutoLock autoLock(lock);
    if (currentBuffer) {
        ScopedBuffer b(*this, data);
    } else {
        currentBuffer = new Buffer(data);
    }
}

void usage() {
    std::cout << "Available options: ";
    for (Options::const_iterator i = options.begin(); i != options.end(); ++i) {
        std::cout << i->first << " ";
    }
    std::cout << std::endl;
}

void showOptions(Options &options) {
    for (Options::const_iterator i = options.begin(); i != options.end(); ++i) {
        std::cout << i->first << "\t" << ((i->second) ? "yes" : "no") << std::endl;
    }
}

void parseOptions(int argc, char** argv, Options &options) {
    for (int i = 1; i < argc; ++i) {
        std::string arg(argv[i]);
        Options::iterator it = options.find(arg);
        if (it != options.end()) {
            options[arg] = true;
        }
        else {
            usage();
            std::stringstream msg;
            msg << "Invalid option: ";
            msg << arg;
            throw std::runtime_error(msg.str());
        }
    }
}

void initOptions(Options &options) {
    options[CUDA_HOSTMEMORY] = false; // cuda kernels use host memory if true (this is less efficient than gpu's memory)
    options[PROFILING] = false;       // enable profiling of cuda operations (to be used with nvprof)
    options[DISABLEVSYNC] = false;    // disable VSYNC for maximum frame rate
}

void showGrabberInfo(MyGrabber &grabber) {
    std::cout << grabber.getWidth() << " x " << grabber.getHeight() << " in " << grabber.getPixelFormat() << std::endl;
}

void checkFormat(MyGrabber &grabber) {
    if (grabber.getPixelFormat() != "Mono8") {
        throw std::runtime_error("This sample only works for Mono8 PixelFormat");
    }
}

int main(int argc, char** argv) {
    int exitCode = 0;
    std::vector<unsigned char*>pinnedMemory(NB_BUFFERS);
    initOptions(options);

    try {
        Euresys::GenTL genTL;
        MyGrabber egrabber(genTL);
        grabber = &egrabber;

        parseOptions(argc, argv, options);
        showOptions(options);

        egrabber.runScript("./config.js");
        showGrabberInfo(egrabber);
        checkFormat(egrabber);

        initOpenGL(&argc, argv);
        initCuda();

        allocateAndAnnounceBuffers(egrabber, pinnedMemory);
        egrabber.start();

        GLuint tex = setupGLObjects(egrabber.getWidth(), egrabber.getHeight());
        setupCudaResources(egrabber.getWidth(), egrabber.getHeight(), tex);

        glutMainLoop();
    }
    catch (const std::exception &e) {
        std::cerr << e.what() << std::endl;
        exitCode = 1;
    }
    catch (...) {
        std::cerr << "uncaught exception!" << std::endl;
        exitCode = 1;
    }

    releaseBuffers(pinnedMemory);
    delete currentBuffer;
    cleanupCuda();

    std::cout << "all done." << std::endl;
    return exitCode;
}
